# gzip_filter.py:  A small demonstration of output filtering.
#
# Copyright (C) 2000, Jon Travis (jtravis@p00p.org)
#
# Description:
#   This module examines headers sent by the client, and determines if gzip
#   encoding is supported.  If it is, input brigades are processed, and sent
#   out in a gzipped fashion.  Notice that this filter is a connection based
#   filter, so all filters changing content will be run before this one.
#
# How to use it:
#   Simply add the following line to your httpd.conf configuration file:
#     SnakeModule gzip_filter.GzipFilter
#   This assumes that SnakeModuleDir has been setup correctly to contain
#   the path that this module resides in.
#
#   The module supports a number of directives which can be used in the main
#   configuration file as well as per-directory .htaccess files.  They are:
#
#   GzipEnable    <on|off> - Enable or disable compressing
#   GzipLevel     <0-9>    - Level of compression. 0 = no compression, 9 = best
#   GzipBlockSize <bsize>  - Optimal size of buckets to pass around.
#                            Default: 8096
#   GzipBuffer    <bsize>  - Number of bytes to potentially buffer up before
#                            flushing the compressor.  Default: 2MB

import gzip
import zlib
import mod_snake
import string
import sys
try:    from cStringIO import StringIO
except: from StringIO import StringIO

FILTERNAME          = 'gzip'
DEFAULT_COMPRESS    = 7
DEFAULT_BLOCKSIZE   = 1024 * 8  
DEFAULT_UNCOMPBYTES = 1 << 20

# Per-dir index
PER_DIR_ENABLED     = 0
PER_DIR_LEVEL       = 1
PER_DIR_BLOCKSIZE   = 2
PER_DIR_UNCOMPBYTES = 3

class WriteLenFile:
    """A class wrapper around StringIO operations, to ensure that
    the length can be had.  cStringIO will not inform the caller about
    the length of the stored data"""
    def __init__(self, str_io_class):
        self.str_io_class = str_io_class
        self.length       = 0
        self.reset()
        
    def reset(self):
        """reset():  Reset the internal state of the file to empty.  All
        data previously stored is discarded"""
        self.str_io = self.str_io_class()
        
    def write(self, data):
        """write(data):  Write data to the file"""
        self.str_io.write(data)
        self.length = self.length + len(data)

    def getvalue(self):
        """getvalue():  Return the contents of the file"""
        return self.str_io.getvalue()

    def flush(self):
        pass
    
class myGzipFile(gzip.GzipFile):
    def __init__(self, *args, **kw):
        apply(gzip.GzipFile.__init__, (self,) + args, kw)

    # Overload write(), because often, the zlib object returns
    # a 0 length string, which is then passed on to write.. Lame.
    def write(self,data):
        if self.fileobj is None:
            raise ValueError, "write() on closed GzipFile object"
        if len(data) > 0:
            self.size = self.size + len(data)
            self.crc = zlib.crc32(data, self.crc)
            comp_data = self.compress.compress(data)
            if len(comp_data):
                self.fileobj.write(comp_data)

class GzipBuffer:
    """Buffer Gzipped data.  When the buffer becomes full, a callback is
    made, and internal data is reset"""

    def __init__(self, gzip_bufsize, level, blocksize, cback_fn, cback_data):
        self.cback_fn   = cback_fn
        self.cback_data = cback_data
        self.blocksize  = blocksize
        self.gzip_data  = WriteLenFile(StringIO)
        self.gzipper    = myGzipFile(None, 'w', level, self.gzip_data)
        self.uncomp_bytes = 0
        self.gzip_bufsize = gzip_bufsize
        
    def __reset_innards(self):
        self.gzip_data.reset()
        self.uncomp_bytes = 0
        
    def flush(self):
        data = self.gzip_data.getvalue()
        for idx in xrange(0, len(data), self.blocksize):
            self.cback_fn(data[idx:idx + self.blocksize], self.cback_data)
        self.__reset_innards()
        
    def write(self, data):
        for idx in xrange(0, len(data), self.blocksize):
            small_chunk = data[idx:idx+self.blocksize]
            self.gzipper.write(small_chunk)
            self.uncomp_bytes = self.uncomp_bytes + len(small_chunk)
            if self.uncomp_bytes > self.gzip_bufsize:
                self.gzipper.flush()
                self.flush()

    def close(self):
        self.gzipper.close()
        self.flush()

class GzipFilter:
    def __init__(self, module):
        module.add_hook('insert_filter',     self.insert_filter)
        module.add_hook('create_dir_config', self.create_dir_config)
        
        module.register_output_filter(FILTERNAME, self.gzip_filter,
                                      mod_snake.AP_FTYPE_CONTENT)
        directives = { "GzipEnable" : (mod_snake.OR_ALL, mod_snake.FLAG,
                                       self.cmd_GzipEnable),
                       "GzipLevel"  : (mod_snake.OR_ALL, mod_snake.TAKE1,
                                       self.cmd_GzipLevel),
                       "GzipBlockSize" : (mod_snake.OR_ALL, mod_snake.TAKE1,
                                          self.cmd_GzipBlockSize),
                       "GzipBuffer" : (mod_snake.OR_ALL, mod_snake.TAKE1,
                                       self.cmd_GzipBuffer) }
        module.add_directives(directives)

    def cmd_GzipEnable(self, per_dir, per_svr, on_or_off):
        per_dir[PER_DIR_ENABLED] = on_or_off

    def cmd_GzipLevel(self, per_dir, per_svr, level):
        """a gzip level between 0 and 9"""
        try:    level = int(level)
        except: return "level must be an integer!"
        if level < 0 or level > 9: 
            return "level must be between 0 and 9"
        per_dir[PER_DIR_LEVEL] = level
        
    def cmd_GzipBlockSize(self, per_dir, per_svr, blocksize):
        """size of blocks, in bytes, to buffer before sending"""
        try:    blocksize = int(blocksize)
        except: return "blocksize must be an integer!"
        if blocksize < 0: return "blocksize must be a positive integer"
        per_dir[PER_DIR_BLOCKSIZE] = blocksize
        
    def cmd_GzipBuffer(self, per_dir, per_svr, buff_size):
        """# of uncompressed bytes to process before flushing"""
        try:    buffsize = int(blocksize)
        except: return "buffsize must be an integer!"
        if blocksize < 0: return "buffsize must be a positive integer"
        per_dir[PER_DIR_UNCOMPBYTES] = buffsize

    def create_dir_config(self, path):
        return [ 0, DEFAULT_COMPRESS, DEFAULT_BLOCKSIZE,
                 DEFAULT_UNCOMPBYTES ]

    def insert_filter(self, per_dir, per_svr, r):
        gzip_buf = GzipBuffer(per_dir[PER_DIR_UNCOMPBYTES],
                              per_dir[PER_DIR_LEVEL],
                              per_dir[PER_DIR_BLOCKSIZE],
                              self.more_buckets_cb, None)
        r.add_output_filter(FILTERNAME, gzip_buf)
        return mod_snake.OK
    
    def more_buckets_cb(self, data, args):
        req, filter = args
        brigade = req.brigade_create()
        brigade.bucket_create_simple(data)
        filter.next.pass_brigade(brigade)
        
    def gzip_filter(self, per_dir, per_svr, filter, brigade):
        req = filter.r
        if (not per_dir[PER_DIR_ENABLED] or
            not req.headers_in.has_key('Accept-Encoding') or
            string.find(req.headers_in['Accept-Encoding'], 'gzip') == -1):
            # Don't process this client -- simply pass it on
            filter.next.pass_brigade(brigade)
            return mod_snake.DECLINED
            
        req.content_encoding = 'gzip'
        del req.headers_out['Content-Length']
        gzipper = filter.ctx
        gzipper.cback_data = [req, filter]
        node = brigade.get_head()
        while node and not node.is_eos:
            length = node.length
            while 1:
                chunk = node.read(1)
                if not chunk:    break
                gzipper.write(chunk)
                if length != -1: break

            node = brigade.get_next(node)

        if node and node.is_eos:
            gzipper.close()
            new_b = req.brigade_create()
            new_b.move_tail(node)
            filter.next.pass_brigade(new_b)

        return mod_snake.OK
