"""
02_hello_world.py: A slightly more complex hello world example.  It may
                   occasionally be convenient for a module to change its
                   behaviour based on which server a request is made on.
                   This module allows the user to use the
                   directive: "HelloWorldURI" in the configuration file
                   to designate which URI returns the "Hello World" message.
                   Different virtual hosts can use the HelloWorldURI directive
                   to set which URI returns the 'Hello World' message on the
                   designated virtual host.

*CONCEPTS:
- Content generation via the content_handler
- Directive creation
- Per-server configuration creation and usage

*TO USE:
Insert the following lines in your httpd.conf file:
----
SnakeModuleDir /path/to/mod_snake/examples/tut
SnakeModule 01_hello_world.Hello_World
----
where /path/to/mod_snake is the path to the extracted mod_snake source.

To demonstrate per-server configuration changes, ensure that more than
1 virtual server is setup, and add HelloWorldURI directives similar to the
following: 

HelloWorldURI /hello_world_main
Listen 8081
<VirtualHost _default_:8081>
HelloWorldURI  /hello_world_8081
</VirtualHost>

When connecting to the main server, getting the URL /hello_world_main,
will return the message 'Hello World', and when connecting to port
8081, as in: http://localhost:8081/hello_world_8081 .. the Hello World
message will be printed.
"""

import mod_snake

class Hello_World:
    def __init__(self, module):
        # First, directives must be setup so that the rest of the configuration
        # file can use them.  The breakdown of the arguments is:
        #
        # "HelloWorldURI" - This is the directive which can be added to the
        #                   configuration file (like ServerName, etc)
        #
        #  mod_snake.RSRC_CONF - This tells Apache that the HelloWorldURI
        #                        directive can appear anywhere within the
        #                        main configuration file, as long as it isn't
        #                        in a <Directory> or <Location> tag
        #
        #  mod_snake.TAKE1     - This tells Apache that the HelloWorldURI
        #                        directive only takes 1 argument in the
        #                        config files
        #
        # Multiple directives can be added to the dictionary before they
        # are added.

        directives = { "HelloWorldURI" : (mod_snake.RSRC_CONF, mod_snake.TAKE1,
                                          self.cmd_HelloWorldURI) }
        module.add_directives(directives)

        # Hook into the 'create_svr_config' callback.  This allows us to
        # create a data structure which will be passed around to our various
        # callbacks.  This callback will be called for the main server config,
        # and virtualhosts
        module.add_hook('create_svr_config', self.create_svr_config)

        # Hook into the content_handler phase.  This allows us to generate
        # content when Apache requests it
        module.add_hook('content_handler',   self.content_handler)
        
    def create_svr_config(self, svr):
        # The job of this routine is to return a configuration object that
        # we can use later on in our other callbacks.  In this case, we
        # store some useful information (the server), and set the default
        # URI for our hello_world content handler.
        return { "server" : svr, 
                 "URI"    : "/hello_world" }

    def cmd_HelloWorldURI(self, per_dir, per_svr, take1_arg):
        # This routine gets called when Apache finds a HelloWorldURI directive
        # in the configuration it is reading.  It calls this function with
        # the argument that it was given (take1_arg), and the configurations
        # that we created.  

        per_svr['URI'] = take1_arg

        # We return None because there was no error.  If there were an error
        # to report, we should return a string error.
        return None
        
    def content_handler(self, per_dir, per_svr, r):
        """Callback which happens when Apache needs to generate content.
        This routine should return DECLINED when it doesn't have anything
        to send, or is called with unknown information, or OK when it has
        sent information to the client"""

        # Ensure that the URI the client is requesting is '/hello_world' and
        # that they are asking for the real thing, not just the HEAD
        if r.uri == per_svr['URI'] and r.method == 'GET':
            # Send out the http_headers
            r.content_type = 'text/html'
            r.send_http_header()
            r.rwrite('Hello world!\n')
            return mod_snake.OK

        return mod_snake.DECLINED

        
