/* ====================================================================
 * Copyright (c) 2001 Jon Travis.  All rights reserved.
 * See the file COPYING for information on usage and redistribution.
 * ====================================================================
 */
#include "mod_snake.h"

#ifdef APACHE13
API_VAR_EXPORT module snake_module;
#else
module AP_MODULE_DECLARE_DATA snake_module;
#endif

static int Python_Loaded = 0;   /* Indicates whether or not it has been
                                   initialized yet.  This goes across
                                   all configs, so we need a global */
static PyThreadState *Main_Threadstate = NULL;
static server_rec *Main_Server = NULL;  /* Main server, so we can serve up
					   information to all the PyMods */
static ap_array_header_t *ModuleDirs; /* Paths to load for all the modules   */
static ap_array_header_t *Modules;    /* Modules loaded (ModSnakePyMod)      */

PyThreadState *mod_snake_main_threadstate(){
  return Main_Threadstate;
}

ap_array_header_t *mod_snake_get_modules(){
  return Modules;
}

/*
 * mod_snake_initialize_python:  Initialize the Python library and setup the
 *                               main thread state.
 */

static void mod_snake_initialize_python(void){
  MS_TRACE;
  if(Python_Loaded) 
    return;

  Python_Loaded = 1;
  Py_Initialize();
#ifdef WITH_THREAD
  PyEval_InitThreads();  /* Acquires a global interp lock for us */
#endif
  Main_Threadstate = PyThreadState_Get();
  PyEval_SaveThread();   /* Release the main thread && Lock      */
}

static void mod_snake_finalize_python(void){
  MS_TRACE;
  if(!Python_Loaded)
    return;

  Python_Loaded = 0;
  PyEval_RestoreThread(Main_Threadstate);   /* Locks   */
  Py_Finalize();                            /* Unlocks */
#ifdef WITH_THREAD
  PyEval_ReleaseLock();
#endif
}


server_rec *mod_snake_get_main_server(){
  return Main_Server;
}


static
void mod_snake_common_post_cfg(ap_pool_t *ap_cfg, ap_pool_t *plog,
			       ap_pool_t *ptemp, server_rec *s)
{
  ModSnakeSCfg *cfg = mod_snake_get_svr_cfg(s);
  char *server_string = PACKAGE "/" VERSION;
  int i;

#ifndef APACHE13
  ap_add_version_component(ap_cfg, server_string);
#else
  ap_add_version_component(server_string);
#endif

  for(i=0; i< Modules->nelts; i++){
    ModSnakePyMod *pymod = (ModSnakePyMod *)Modules->elts + i;

    if(!pymod->hooks[SNAKE_HOOK_POST_CONFIG]) continue;
    mod_snake_run_hook_post_config(pymod, cfg);
  }
}

/*
 * mod_snake_post_cfg_cb:  Post configuration callback.  Setup per-module
 *                         type things such as thread locks.
 */

#ifndef APACHE13
static 
void mod_snake_post_cfg_cb(ap_pool_t *p, ap_pool_t *plog,
                           ap_pool_t *ptemp, server_rec *s)
{
  mod_snake_common_post_cfg(p, plog, ptemp, s);
}
#endif

/*
 * unload_dir_cfg:  Unload a per-directory configuration entity.  The per-dir
 *                  config contains callback data for all modules which have
 *                  a registered directive underneath that directory.
 *
 * Arguments:       vdata = ModSnakeDCfg directory configuration.
 *
 *
 * Side Notes:      XXX -- If the directory config is destroyed AFTER the
 *                         server it was created within, then some data
 *                         will be orphaned.  A better explaination:  The
 *                         module thread + interp data is stored at the server
 *                         config level.  When directory configs are created,
 *                         the callback data is really associated with these
 *                         interps within the servercfg.  Therefore when a
 *                         servercfg destroys the interp, and then the
 *                         dir_cfg wants to destroy data that it kept from
 *                         that interp, it can't because it has been orphaned.
 *                         We could probably get away with decrefing these in
 *                         the main interpreter.
 */
static
#ifdef APACHE13
void 
#else
ap_status_t
#endif
unload_dir_cfg(void *vdata){
  ModSnakeDCfg    *cfg         = vdata;
  ap_array_header_t *dir_data    = cfg->dir_config_data;
  int i;

  MS_TRACE;
  for(i=0; i< dir_data->nelts; i++){
    ModSnakeCallbackData *cdata = (ModSnakeCallbackData *)dir_data->elts + i;

    /* Un-threadsafe? - XXX */
    if(cdata->module->main_state == NULL){
      fprintf(stderr, "Orphaned directory config data in module: %p\n", 
              cdata->module);
      continue;
    }
    PyEval_RestoreThread(cdata->module->main_state);       /* Lock */
    Py_DECREF(cdata->data);
    PyEval_SaveThread();                                   /* Unlock */
  }
#ifdef APACHE13
  return;
#else
  return APR_SUCCESS;
#endif
}

/*
 * unload_svr_cfg:  Unload a server configuration entity.  All modules
 *                  created within this server entity will be unloaded, and
 *                  all other Python objects within this server will be 
 *                  decref'd.
 *
 * Arguments:       vdata = Server configuration entity
 */

static
#ifdef APACHE13
void 
#else
ap_status_t
#endif
unload_svr_cfg(void *vdata){
  ModSnakeSCfg *cfg         = vdata;
  ap_array_header_t *svr_data    = cfg->svr_config_data;
  int i;

  MS_TRACE;
  for(i=0; i< svr_data->nelts; i++){
    ModSnakeCallbackData *cdata = (ModSnakeCallbackData *)svr_data->elts + i;

    /* Un-threadsafe? - XXX */
    if(cdata->module->main_state == NULL){
      fprintf(stderr, "Orphaned server config data in module: %p\n", 
              cdata->module);
      continue;
    }
    PyEval_RestoreThread(cdata->module->main_state);       /* Lock */
    Py_DECREF(cdata->data);
    PyEval_SaveThread();                                   /* Unlock */
  }

#ifdef APACHE13
  return;
#else
  return APR_SUCCESS;
#endif
}

static AP_CLEANUP_RESTYPE cleanup_special_global(void *vdata){
  MS_TRACE;
  ModuleDirs  = NULL;
  Modules     = NULL;
  Main_Server = NULL;
  mod_snake_finalize_python();  /* Finalize all the interps, etc */
  return AP_CLEANUP_RESULT(0);
}

static void setup_special_global(ap_pool_t *p, server_rec *ap_main_svr){
  MS_TRACE;
  
  if(ModuleDirs != NULL)
    fprintf(stderr, "*Warning* Initializing main server twice!\n");

  ModuleDirs  = ap_make_array(p, 50, sizeof(char *));
  Modules     = ap_make_array(p, 50, sizeof(ModSnakePyMod));
  Main_Server = ap_main_svr;
  mod_snake_initialize_python();  /* Ensure that we are setup */
  ap_register_cleanup(p, NULL, cleanup_special_global, ap_null_cleanup);
}

static void *mod_snake_create_svr_cfg_cb(ap_pool_t *p, server_rec *s){
  ModSnakeSCfg *res = ap_pcalloc(p, sizeof(*res));

  MS_TRACE;
  res->cfgpool         = p;
  res->server          = s;
  res->parent          = NULL;
  res->svr_config_data = ap_make_array(p, 50, sizeof(ModSnakeCallbackData));
  ap_register_cleanup(p, res, unload_svr_cfg, ap_null_cleanup);

  /* Main server configuration -- do special setup of global vars */
  if(s->defn_name == NULL)
    setup_special_global(p, s);
  return res;
}

static void *mod_snake_create_dir_cfg_cb(ap_pool_t *p, char *dir){
  ModSnakeDCfg *res;

  MS_TRACE;
  res = ap_pcalloc(p, sizeof(*res));
  res->cfgpool         = p;
  res->dir_config_data = ap_make_array(p, 50, sizeof(ModSnakeCallbackData));
  res->path            = ap_pstrdup(p, dir);
  res->parent          = NULL;
  ap_register_cleanup(p, res, unload_dir_cfg, ap_null_cleanup);
  return res;
}

/*
 * mod_snake_merge_dir_cfg_cb:  Callback for merging of per-dir configuration
 *                              data.  Here, we loop through the modules
 *                              involved in the merge.  Those that want
 *                              to be merged (i.e. they have hooked the
 *                              handler), we call the handler, and store their
 *                              result.  Those who do not want to be merged,
 *                              we simply add them to this new dir-cfg and
 *                              incref their respective callback data.
 */

static void *mod_snake_merge_cfg_cb(ap_pool_t *p, void *base_vdata,
				    void *newc_vdata, int cfg_hook)
{
  ModSnakeDCfg *dbase = base_vdata, *dnewcfg = newc_vdata, *dres = NULL;
  ModSnakeSCfg *sbase = base_vdata, *snewcfg = newc_vdata, *sres = NULL;
  ap_array_header_t *cfg_base, *cfg_new, *res_cfg_data;
  int i, is_dir = (cfg_hook == SNAKE_HOOK_MERGE_DIR_CONFIG);

  MS_TRACE;
  if(is_dir) {
    dres         = mod_snake_create_dir_cfg_cb(p, dnewcfg->path);
    dres->parent = dbase;
    cfg_base     = dbase->dir_config_data;
    cfg_new      = dnewcfg->dir_config_data;
    res_cfg_data = dres->dir_config_data;
  } else {
    sres         = mod_snake_create_svr_cfg_cb(p, snewcfg->server);
    sres->parent = sbase;
    cfg_base     = sbase->svr_config_data;
    cfg_new      = snewcfg->svr_config_data;
    res_cfg_data = sres->svr_config_data;
  }

  for(i=0; i < cfg_base->nelts; i++){
    ModSnakeCallbackData *bdata = (ModSnakeCallbackData *)cfg_base->elts + i;
    int j;
    
    for(j=0; j < cfg_new->nelts; j++){
      ModSnakeCallbackData *ndata = (ModSnakeCallbackData *)cfg_new->elts + j;
      ModSnakeCallbackData *new_resdata;

      if(ndata->module == bdata->module){
        PyObject *hook, *data;
        ModSnakePyMod *pymod = ndata->module;

        if((hook = pymod->hooks[cfg_hook]) == NULL){
          data = ndata->data;
          Py_INCREF(data);
        } else {

	  data = (is_dir ? mod_snake_run_hook_merge_dir_config : 
		  mod_snake_run_hook_merge_svr_config)(pymod, bdata->data,
						       ndata->data, hook);
          if(data == NULL){
            fprintf(stderr, "Error calling merge_%s for: %s\n",
                    is_dir ? "dir" : "svr", pymod->name);
            continue;
          }
        }
        new_resdata         = ap_push_array(res_cfg_data);
        new_resdata->data   = data;
        new_resdata->module = pymod;
      }
    }      
  }
  return is_dir ? (void *)dres : (void *)sres;
}


static void *mod_snake_merge_dir_cfg_cb(ap_pool_t *p, void *base, void *newc){
  return mod_snake_merge_cfg_cb(p, base, newc, SNAKE_HOOK_MERGE_DIR_CONFIG);
}

static void *mod_snake_merge_svr_cfg_cb(ap_pool_t *p, void *base, 
					   void *newc)
{
  return mod_snake_merge_cfg_cb(p, base, newc, SNAKE_HOOK_MERGE_SVR_CONFIG);
}

#ifdef APACHE13
static int file_exists(const char *path, ap_pool_t *p){
  struct stat statbuf;
  int err;

  if((err = stat(path, &statbuf)) == -1)
    return -1;
  
  if(S_ISDIR(statbuf.st_mode))
    return -2;

  return 0;
}
#else
static int file_exists(const char *path, ap_pool_t *p){
  ap_finfo_t finfo;
  ap_status_t err;

  if((err = ap_stat(&finfo, path, APR_FINFO_TYPE, p)) != APR_SUCCESS)
    return -1;
  
  if(finfo.filetype == APR_DIR)
    return -2;

  return 0;
}
#endif

/*
 * mod_snake_add_moduledir:  Add a directory to the list of directories
 *                           that contain Python modules for import.
 */

static 
const char *mod_snake_add_moduledir(cmd_parms *cmd, void *dcfg, char *arg) {
  const char *real_path, *invalid;
  int stres;

  MS_TRACE;
  if((invalid = ap_check_cmd_context(cmd, GLOBAL_ONLY)) != NULL) 
    return invalid;

  /* Validate the existance first, so we can give better error msgs */
  real_path = ap_server_root_relative(cmd->pool, arg);
  if((stres = file_exists(real_path, cmd->pool)) == -1)
    return ap_psprintf(cmd->pool, "mod_snake: path '%s' unaccessable", arg);

  if(stres != -2)
    return ap_psprintf(cmd->pool, "mod_snake: path '%s': not a directory", 
                       arg);

  *(char **)ap_push_array(ModuleDirs) = ap_pstrdup(cmd->pool, real_path);
  return NULL;
}

/*
 * mod_snake_find_module_byname:  Find a module by it's module.classname name.
 *                                
 * Arguments:                     mods = Array of ModSnakePyMod's
 *                                name = module.classname to find.
 *
 * Return values:                 Returns NULL if the module was not found,
 *                                else a pointer to it.
 */

ModSnakePyMod *mod_snake_find_module_byname(ap_array_header_t *mods, 
                                            const char *name)
{
  int i;

  for(i=0; i < mods->nelts; i++){
    ModSnakePyMod *pymod = (ModSnakePyMod *)mods->elts + i;
    
    if(!strcmp(pymod->name, name)) return pymod;
  }

  return NULL;
}

/*
 * mod_snake_get_cback_data:  Check a callback config to see if a 
 *                            module has had its create_{dir,svr}_config hook
 *                            called, and subsequent data stored.  If it
 *                            hasn't, call the module's create_{dir,svr}_config
 *                            hook and store their resultant data.
 *
 * Arguments:                 cback_cfg = Directory or server config to check
 *                            pymod     = Module to check for callback-fu
 *                            s         = Current server in callback
 *                            do_init   = 1 if we need to initialize the
 *                                        modules callback data, regardless
 *                                        if they have a hook or not, 0 
 *                                        otherwise.  (The callback data
 *                                        will be set to 'None' and INCREFd)
 *                            cback_type = One of SNAKE_HOOK_CREATE_DIR_CONFIG
 *                                             or SNAKE_HOOK_CREATE_SVR_CONFIG
 *                             
 *
 * Side Notes:                The interp lock must NOT be held when
 *                            calling this function.
 *
 * Return values:             Returns -1 on an anomaly, else 0
 */

static
int mod_snake_get_cback_data(void *cback_cfg, ModSnakePyMod *pymod,
			     server_rec *s, int do_init, int cback_type)
{
  ap_array_header_t *cback_data;
  ModSnakeCallbackData *cdata;
  PyThreadState *tstate;
  PyObject *data, *hook;

  if(cback_type == SNAKE_HOOK_CREATE_DIR_CONFIG) 
    cback_data = ((ModSnakeDCfg *)cback_cfg)->dir_config_data;
  else
    cback_data = ((ModSnakeSCfg *)cback_cfg)->svr_config_data;

  if((hook = pymod->hooks[cback_type]) == NULL){
    if(do_init){
      if((tstate = mod_snake_create_thread_and_lock(pymod->interp)) == NULL)
	return -1;

      cdata         = ap_push_array(cback_data);
      cdata->data   = Py_None;
      Py_INCREF(cdata->data);
      cdata->module = pymod;
      mod_snake_destroy_thread_and_unlock(tstate);
    }
    return 0;
  }
  
  if(mod_snake_find_cback_data(pymod, cback_data))
    return 0;  /* Already have callback data for this module */

  /* data returned is INCREF'd */
  if(cback_type == SNAKE_HOOK_CREATE_DIR_CONFIG)
    data = mod_snake_run_hook_create_dir_config(pymod, cback_cfg, hook);
  else
    data = mod_snake_run_hook_create_svr_config(pymod, s, cback_cfg, hook);

  if(data == NULL)
    return -1;

  cdata         = ap_push_array(cback_data);
  cdata->data   = data;
  cdata->module = pymod;
  return 0;
}

static AP_CLEANUP_RESTYPE mod_snake_pymod_cleanup(void *vdata){
  ModSnakePyMod *pymod = vdata;
  PyThreadState *oldtstate;
  int i, j;

  MS_TRACE;

  if(pymod->valid){
#ifdef WITH_THREAD
    PyEval_AcquireLock();                     /* Locks */
#endif
    oldtstate = PyThreadState_Swap(pymod->main_state);
    Py_XDECREF(pymod->module);
    Py_XDECREF(pymod->mod_obj);
    Py_XDECREF(pymod->profile_info.profile_obj);
    Py_XDECREF(pymod->profile_info.profile_func);
    for(i=0; i< SNAKE_HOOK_NUM_HOOKS; i++){
      Py_XDECREF(pymod->hooks[i]);
      pymod->hooks[i] = NULL;
    }

    for(j=0; j < SNAKE_FILTERS_NTYPES; j++){
      for(i=0; i< SNAKE_FILTERS_MAX; i++){
	Py_XDECREF(pymod->filters[j][i].func);
	pymod->filters[j][i].func = NULL;
      }
    }

    Py_EndInterpreter(pymod->main_state);
    PyThreadState_Swap(oldtstate);
#ifdef WITH_THREAD
    PyEval_ReleaseLock();                     /* Unlock */
#endif
    pymod->valid = 0;
  }
  return AP_CLEANUP_RESULT(0);
}
    
static 
ModSnakePyMod *mod_snake_pymod_new(ap_array_header_t *modulelist,
				   const char *module_name, int is_profile)
{
  ModSnakePyMod *pymod  = ap_push_array(modulelist);
  ap_pool_t *ap_p = AP_ARRAY_POOL(modulelist);
  int i, j;

  MS_TRACE;

  pymod->ap_p                      = ap_p;
  pymod->name                      = ap_pstrdup(ap_p, module_name);
  pymod->module                    = NULL;  
  pymod->mod_obj                   = NULL;
  pymod->interp                    = NULL;
  pymod->main_state                = NULL;
  pymod->valid                     = 0;
  pymod->profile_info.profile      = is_profile;
  pymod->profile_info.profile_obj  = NULL;
  pymod->profile_info.profile_func = NULL;
  pymod->current_hook              = SNAKE_HOOK_NO_HOOK;
  for(i=0; i< SNAKE_HOOK_NUM_HOOKS; i++)
    pymod->hooks[i] = NULL;

  for(j=0; j < SNAKE_FILTERS_NTYPES; j++){
    for(i=0; i< SNAKE_FILTERS_MAX; i++){
      pymod->filters[j][i].func = NULL;
    }
  }

  ap_register_cleanup(ap_p, pymod, mod_snake_pymod_cleanup, ap_null_cleanup);
  return pymod;
}


/*
 * mod_snake_add_module:  Add a module to the list of modules that mod_snake
 *                        will manage.  The module will then be loaded so
 *                        that we can get hooks and determine if the server
 *                        is even going to run.
 *
 * Arguments:             arg1    = Module name to import (as in import foo)
 *                        arg2..n = (optional)  
 *                          profile = Enables profiling of calls out to this
 *                                    module.  Profiling data is available
 *                                    through API calls or mod_snake_info
 *
 */

static 
const char *mod_snake_add_module(cmd_parms *cmd, void *dcfg, const char *args)
{
  ModSnakeSCfg *svr_cfg = mod_snake_get_svr_cfg(cmd->server);
  ModSnakeDCfg *dir_cfg    = dcfg;
  ModSnakePyMod *pymod;
  char *module_name, *cp;
  const char *invalid;
  int is_profile;

  MS_TRACE;

  if((invalid = ap_check_cmd_context(cmd, GLOBAL_ONLY)) != NULL) 
    return invalid;

  if((module_name = ap_getword_conf(cmd->pool, &args)) == NULL)
    return "mod_snake: First argument must be a module name";

  /* Check to see if this module is already handled within our servercfg */
  if(mod_snake_find_module_byname(Modules, module_name) != NULL)
    return "mod_snake: Duplicate module additions within server not allowed";

  /* Check validity of the module.classname */
  if((cp = strchr(module_name, '.')) == NULL)
    return "mod_snake: Module name must be in the form of module.classname";

  is_profile = 0;
  while(*args){
    char *cp = ap_getword_conf(cmd->pool, &args);
    if(!strcmp(cp, "profile")) is_profile = 1;
    else return "mod_snake: Optional parameter must be 'profile'";
  }

  pymod = mod_snake_pymod_new(Modules, module_name, is_profile);
  
  /* Called pre-fork, pre-thread, so we shoudln't need to worry about doing
     locking .. but oh well */
  PyEval_RestoreThread(Main_Threadstate); /* Locks */
  if(mod_snake_load_module(pymod, ModuleDirs, cmd->server) == -1){
    PyEval_SaveThread();                  /* Unlocks */
    return ap_psprintf(cmd->pool, "mod_snake: Unable to load module %s",
                       module_name);
  }
  PyEval_SaveThread();                    /* Unlocks */

  /* Create the initial per-svr configuration for the module */
  if(mod_snake_get_cback_data(svr_cfg, pymod, cmd->server, 1, 
			      SNAKE_HOOK_CREATE_SVR_CONFIG) == -1) 
    return ap_psprintf(cmd->pool, "mod_snake: Error getting server info "
                       "from module %s", module_name);

  /* Create the initial per-dir configuration for the module */
  if(mod_snake_get_cback_data(dir_cfg, pymod, cmd->server, 1, 
			      SNAKE_HOOK_CREATE_DIR_CONFIG) == -1) 
    return ap_psprintf(cmd->pool, "mod_snake: Error getting directory info "
                       "from module %s", module_name);
  return NULL;
}

static
void mod_snake_open_logs_cb(ap_pool_t *pconf, ap_pool_t *plog, ap_pool_t *ptmp,
                            server_rec *s)
{
  ModSnakeSCfg *cfg = mod_snake_get_svr_cfg(s);
  int i;
  
  for(i=0; i< Modules->nelts; i++){
    ModSnakePyMod *pymod = (ModSnakePyMod *)Modules->elts + i;

    if(!pymod->hooks[SNAKE_HOOK_OPEN_LOGS]) continue;
    mod_snake_run_hook_open_logs(pymod, cfg);
  }
}

#ifdef APACHE13
static void mod_snake_module_init(server_rec *s, pool *p) {
  mod_snake_open_logs_cb(p, p, p, s);
  mod_snake_common_post_cfg(p, p, p, s);
}
#endif /* APACHE13 */

#ifndef APACHE13
static 
int mod_snake_generic_connection_cb(conn_rec *c, int hook_number){
  ModSnakeSCfg *cfg = mod_snake_get_svr_cfg(c->base_server), *svrsrch;
  PyObject *svr_cbdata = NULL;
  int i, res, hook_type;
  
  hook_type = mod_snake_hooknum_to_type(hook_number);
  for(i=0; i< Modules->nelts; i++){
    ModSnakePyMod *pymod = (ModSnakePyMod *)Modules->elts + i;
    PyObject *callback = pymod->hooks[hook_number];

    if(!callback) continue;

    svrsrch = cfg;
    while(svrsrch != NULL){
      if((svr_cbdata = mod_snake_find_svr_cback_data(pymod, svrsrch)) != NULL)
	break;

      svrsrch = svrsrch->parent;
    }

    if(svrsrch == NULL){
      mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0,
			  c->base_server, 
			  ap_pstrcat(c->pool, 
				     "mod_snake: Failed to locate config "
				     "information for module: %s",
				     pymod->name, NULL));
      return HTTP_INTERNAL_SERVER_ERROR;
    }

    res = mod_snake_run_generic_conn_rec(pymod, c, svr_cbdata, callback, 
					 hook_number);
    
    if(hook_type == SNAKE_HOOK_ALL && (res != DECLINED && res != OK))
      return res;

    if(hook_type == SNAKE_HOOK_FIRST && res != DECLINED)
      return res;
  }

  return DECLINED;
}

static 
int mod_snake_pre_connection_cb(conn_rec *c){
  return mod_snake_generic_connection_cb(c, SNAKE_HOOK_PRE_CONNECTION);
}

static 
int mod_snake_process_connection_cb(conn_rec *c){
  return mod_snake_generic_connection_cb(c, SNAKE_HOOK_PROCESS_CONNECTION);
}
#endif

static
int mod_snake_generic_request_cb(request_rec *r, int hook_number)
{
  ModSnakeSCfg *cfg=mod_snake_get_svr_cfg(r->server), *svrsrch;
  ModSnakeDCfg *dcfg = ap_get_module_config(r->per_dir_config, &snake_module),
               *dirsrch;
  PyObject *dir_cbdata = NULL, *svr_cbdata = NULL;
  int i, res = DECLINED, hook_type;
  
  hook_type = mod_snake_hooknum_to_type(hook_number);
  for(i=0; i< Modules->nelts; i++){
    ModSnakePyMod *pymod = (ModSnakePyMod *)Modules->elts + i;
    PyObject *callback = pymod->hooks[hook_number];

    if(!callback) continue;

    /* Unfortunately mod_snake must play the role of Apache all too many
       times in order to proxy the correct information out to all of the
       PyMods.  In this case, we search back through the hierarchy of
       per-{dir,svr} configs, until we find one belonging to this module. At
       the worst case, this should be the main 'server-context' per-*-cfg */
    dirsrch = dcfg;
    while(dirsrch != NULL){
      if((dir_cbdata = mod_snake_find_dir_cback_data(pymod, dirsrch)) != NULL)
	break;

      dirsrch = dirsrch->parent;
    }

    svrsrch = cfg;
    while(svrsrch != NULL){
      if((svr_cbdata = mod_snake_find_svr_cback_data(pymod, svrsrch)) != NULL)
	break;

      svrsrch = svrsrch->parent;
    }

    if(svrsrch == NULL || dirsrch == NULL){
      mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0,
			  r->server, 
			  ap_pstrcat(r->pool, 
				     "mod_snake: Failed to locate config "
				     "information for module: %s",
				     pymod->name, NULL));
      return HTTP_INTERNAL_SERVER_ERROR;
    }

    res = mod_snake_run_generic_request_rec(pymod, r, dir_cbdata, svr_cbdata,
					    callback, hook_number);

    if(hook_type == SNAKE_HOOK_ALL && (res != DECLINED && res != OK))
      return res;

    if(hook_type == SNAKE_HOOK_FIRST && res != DECLINED)
      return res;
  }
  return DECLINED;
}

static 
int mod_snake_header_parser_cb(request_rec *r){
  return mod_snake_generic_request_cb(r, SNAKE_HOOK_HEADER_PARSER);
}

static 
int mod_snake_post_read_request_cb(request_rec *r){
  return mod_snake_generic_request_cb(r, SNAKE_HOOK_POST_READ_REQUEST);
}

static 
int mod_snake_log_transaction_cb(request_rec *r){
  return mod_snake_generic_request_cb(r, SNAKE_HOOK_LOG_TRANSACTION);
}

static 
int mod_snake_translate_name_cb(request_rec *r){
  return mod_snake_generic_request_cb(r, SNAKE_HOOK_TRANSLATE_NAME);
}

static 
int mod_snake_check_user_id_cb(request_rec *r){
  return mod_snake_generic_request_cb(r, SNAKE_HOOK_CHECK_USER_ID);
}
static 
int mod_snake_fixups_cb(request_rec *r){
  return mod_snake_generic_request_cb(r, SNAKE_HOOK_FIXUPS);
}

static 
int mod_snake_type_checker_cb(request_rec *r){
  return mod_snake_generic_request_cb(r, SNAKE_HOOK_TYPE_CHECKER);
}

static 
int mod_snake_access_checker_cb(request_rec *r){
  return mod_snake_generic_request_cb(r, SNAKE_HOOK_ACCESS_CHECKER);
}

static 
int mod_snake_auth_checker_cb(request_rec *r){
  return mod_snake_generic_request_cb(r, SNAKE_HOOK_AUTH_CHECKER);
}

static
int mod_snake_content_handler_cb(request_rec *r){
  return mod_snake_generic_request_cb(r, SNAKE_HOOK_CONTENT_HANDLER);
}

#ifndef APACHE13
static void mod_snake_insert_filter_cb(request_rec *r){
  mod_snake_generic_request_cb(r, SNAKE_HOOK_INSERT_FILTER);
}

static void mod_snake_register_hooks_cb(ap_pool_t *p)
{
  ap_hook_open_logs(mod_snake_open_logs_cb, NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_post_config(mod_snake_post_cfg_cb, NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_pre_connection(mod_snake_pre_connection_cb, NULL, NULL, 
                         APR_HOOK_MIDDLE);
  ap_hook_process_connection(mod_snake_process_connection_cb, NULL, NULL, 
                             APR_HOOK_MIDDLE);
  ap_hook_post_read_request(mod_snake_post_read_request_cb, NULL, 
                            NULL, APR_HOOK_MIDDLE);
  ap_hook_header_parser(mod_snake_header_parser_cb, NULL, NULL,APR_HOOK_MIDDLE);
  ap_hook_log_transaction(mod_snake_log_transaction_cb, NULL, NULL,
                          APR_HOOK_MIDDLE);
  ap_hook_translate_name(mod_snake_translate_name_cb, NULL, NULL,
                         APR_HOOK_MIDDLE);
  ap_hook_check_user_id(mod_snake_check_user_id_cb, NULL, NULL,
                        APR_HOOK_MIDDLE);
  ap_hook_fixups(mod_snake_fixups_cb, NULL, NULL,
                 APR_HOOK_MIDDLE);
  ap_hook_type_checker(mod_snake_type_checker_cb, NULL, NULL,
                       APR_HOOK_MIDDLE);
  ap_hook_access_checker(mod_snake_access_checker_cb, NULL, NULL,
                         APR_HOOK_MIDDLE);
  ap_hook_auth_checker(mod_snake_auth_checker_cb, NULL, NULL,
                       APR_HOOK_MIDDLE);
  ap_hook_handler(mod_snake_content_handler_cb, NULL, NULL,
		  APR_HOOK_MIDDLE);
  ap_hook_insert_filter(mod_snake_insert_filter_cb, NULL, NULL,
			APR_HOOK_MIDDLE);
}
#endif

static const command_rec mod_snake_cmds[] = {
{ "SnakeModuleDir", mod_snake_add_moduledir, NULL, RSRC_CONF, ITERATE,
  "Load a module to be used by mod_snake" },
{ "SnakeModule",    mod_snake_add_module,    NULL, RSRC_CONF, RAW_ARGS,
  "Load a module to be used by mod_snake : \
   SnakeModule module.moduleclass [profile]"},
{ NULL }
};

#ifndef APACHE13
module AP_MODULE_DECLARE_DATA snake_module = {
  STANDARD20_MODULE_STUFF,
  mod_snake_create_dir_cfg_cb,    /* create per-directory config structure */
  mod_snake_merge_dir_cfg_cb,     /* merge per-directory config structure  */
  mod_snake_create_svr_cfg_cb,    /* create per-server config structure    */
  mod_snake_merge_svr_cfg_cb,     /* merge per-server config structure     */
  mod_snake_cmds,                 /* command ap_table_t                    */
  mod_snake_register_hooks_cb,    /* register hooks                        */
};
#endif

#ifdef APACHE13
static handler_rec mod_snake_handlers [] = {
  { "*",            mod_snake_content_handler_cb },
  { DIR_MAGIC_TYPE, mod_snake_content_handler_cb },
  { NULL },
};

API_VAR_EXPORT module snake_module = {
  STANDARD_MODULE_STUFF,
  mod_snake_module_init,          /* module initializer */
  mod_snake_create_dir_cfg_cb,    /* per-directory config creator */
  mod_snake_merge_dir_cfg_cb,     /* dir config merger */
  mod_snake_create_svr_cfg_cb,    /* server config creator */
  mod_snake_merge_svr_cfg_cb,     /* server config merger */
  mod_snake_cmds,                 /* command table */
  mod_snake_handlers,             /* [7] list of handlers */
  mod_snake_translate_name_cb,    /* [2] filename-to-URI translation */
  mod_snake_check_user_id_cb,     /* [5] check/validate user_id */
  mod_snake_auth_checker_cb,      /* [6] check user_id is valid *here* */
  mod_snake_access_checker_cb,    /* [4] check access by host address */
  mod_snake_type_checker_cb,      /* [7] MIME type checker/setter */
  mod_snake_fixups_cb,            /* [8] fixups */
  mod_snake_log_transaction_cb,   /* [10] logger */
  mod_snake_header_parser_cb,     /* [3] header parser */
  NULL,                           /* apache child process initializer */
  NULL,                           /* apache child process exit/cleanup */
  mod_snake_post_read_request_cb  /* [1] post read_request handling */
};                                                    
#endif

ModSnakeSCfg *mod_snake_get_svr_cfg(server_rec *svr){
  return ap_get_module_config(svr->module_config, &snake_module);
}

ModSnakeDCfg *mod_snake_get_dir_cfg(request_rec *req){
  return ap_get_module_config(req->per_dir_config, &snake_module);
}

static void init_common_conn_cfg(ModSnakeConnCfg *cfg){
  cfg->pymod = NULL;
}

static AP_CLEANUP_RESTYPE reqcfg_notes_cleanup(void *vdata){
  PyObject *pdict = vdata;

  PyEval_RestoreThread(mod_snake_main_threadstate()); /* Locks */
  Py_DECREF(pdict);
  PyEval_SaveThread();
  return AP_CLEANUP_RESULT(0);
}

ModSnakeReqCfg *mod_snake_get_req_cfg(request_rec *req){
  ModSnakeReqCfg *res = ap_get_module_config(req->request_config, 
					     &snake_module);

  /* Setup the request config if it doesn't yet exist */
  if(!res){
    res = ap_pcalloc(req->pool, sizeof(*res));
    if((res->notes = PyDict_New()) == NULL)
      return NULL;
#ifdef APACHE13
    init_common_conn_cfg(res);
#else
    res->curbuf  = NULL;
    res->bufused = 0;
#endif
    ap_register_cleanup(req->pool, res->notes, reqcfg_notes_cleanup,
			ap_null_cleanup);
    ap_set_module_config(req->request_config, &snake_module, res);
  }
  return res;
}

ModSnakeConnCfg *mod_snake_get_conn_cfg(conn_rec *conn, request_rec *req){
#ifdef APACHE13
  return mod_snake_get_req_cfg(req);
#else
  ModSnakeConnCfg *res = ap_get_module_config(conn->conn_config,&snake_module);
  if(!res){
    res = ap_pcalloc(conn->pool, sizeof(*res));
    init_common_conn_cfg(res);
    ap_set_module_config(conn->conn_config, &snake_module, res);
  }
  return res;
#endif
}

static 
const char *mod_snake_setup_internal_directive_call(cmd_parms *cmd,
                                                   void *per_dir_vdata,
                                                   PyThreadState **new_state,
                                                   PyObject **callback_dirdata,
                                                   PyObject **callback_svrdata,
                                                   PyObject **callback_func,
                                                   ModSnakePyMod **pymod)
{
  ModSnakeSCfg *svrcfg = mod_snake_get_svr_cfg(cmd->server);
  ModSnakeDCfg *dircfg = per_dir_vdata;
  ModSnakeIntDirective *int_dir;

  if(!(int_dir = cmd->cmd->cmd_data))
    return "mod_snake:  Object handling command has disappeared";

  if(mod_snake_get_cback_data(svrcfg, int_dir->pymod, cmd->server, 0,
			      SNAKE_HOOK_CREATE_SVR_CONFIG) == -1)
    return "mod_snake:  Error setting up per-svr config data!";

  if(mod_snake_get_cback_data(per_dir_vdata, int_dir->pymod, cmd->server, 0,
			      SNAKE_HOOK_CREATE_DIR_CONFIG) == -1)
    return "mod_snake:  Error setting up per-dir config data!";

  *pymod = int_dir->pymod;

  /* Setup a new threadstate for us to have a working context for calling */
  /* Lock */
  *new_state = mod_snake_create_thread_and_lock((*pymod)->interp);
  if(*new_state == NULL)
    return "mod_snake:  Failed to create a new interpreter state!";

  /* Find per-svr config data to send to the callback */
  if(!(*callback_svrdata = mod_snake_find_svr_cback_data(*pymod, svrcfg)))
    *callback_svrdata = Py_None;

  /* Find per-dir config data to send to the callback */
  if(!(*callback_dirdata = mod_snake_find_dir_cback_data(*pymod, dircfg)))
    *callback_dirdata = Py_None;

  *callback_func = int_dir->funcptr;
  return NULL;
}

static const 
char *mod_snake_internal_directive_no_args(cmd_parms *cmd, void *per_dir_vdata)
{
  ModSnakePyMod *pymod;
  PyThreadState *my_state;
  PyObject *func, *dir_cbdata, *svr_cbdata, *args, *res;
  const char *cres;

  /* Lock */
  cres = mod_snake_setup_internal_directive_call(cmd, per_dir_vdata, &my_state,
						 &dir_cbdata, &svr_cbdata, 
						 &func, &pymod);
  if(cres) return cres;

  args = Py_BuildValue("(OO)", dir_cbdata, svr_cbdata);  /* Incref */
  res = mod_snake_pyeval_callobject(pymod, func, args, SNAKE_HOOK_DIRECTIVE);
  Py_DECREF(args);
  if(res == NULL){
    if(PyErr_Occurred()) PyErr_Print();
    mod_snake_destroy_thread_and_unlock(my_state);  /* Unlock */
    return ap_psprintf(cmd->pool, "mod_snake: Error calling command '%s'",
                      cmd->cmd->name);
  }

  cres = PyString_Check(res) ? ap_pstrdup(cmd->pool, PyString_AsString(res))
                             : NULL;

  Py_DECREF(res);
  mod_snake_destroy_thread_and_unlock(my_state);  /* Unlock */
  return cres;
}

static const 
char *mod_snake_internal_directive_take1(cmd_parms *cmd, void *per_dir_vdata,
                                         const char *arg1)
{
  ModSnakePyMod *pymod;
  PyThreadState *my_state;
  PyObject *func, *dir_cbdata, *svr_cbdata, *args, *res;
  const char *cres;

  /* Lock */
  cres = mod_snake_setup_internal_directive_call(cmd, per_dir_vdata, &my_state,
						 &dir_cbdata, &svr_cbdata,
						 &func, &pymod);
  if(cres) return cres;

  if(cmd->cmd->args_how == FLAG)
    args = Py_BuildValue("(OOi)", dir_cbdata, svr_cbdata, (int)arg1); /* INC */
  else
    args = Py_BuildValue("(OOs)", dir_cbdata, svr_cbdata, arg1);      /* INC */

  res = mod_snake_pyeval_callobject(pymod, func, args, SNAKE_HOOK_DIRECTIVE);
  Py_DECREF(args);
  if(res == NULL){
    if(PyErr_Occurred()) PyErr_Print();
    mod_snake_destroy_thread_and_unlock(my_state);  /* Unlock */
    return ap_psprintf(cmd->pool, "mod_snake: Error calling command '%s'",
                      cmd->cmd->name);
  }

  cres = PyString_Check(res) ? ap_pstrdup(cmd->pool, PyString_AsString(res))
                             : NULL;
  Py_DECREF(res);
  mod_snake_destroy_thread_and_unlock(my_state);  /* Unlock */
  return cres;
}

static const 
char *mod_snake_internal_directive_take2(cmd_parms *cmd, void *per_dir_vdata,
                                         const char *arg1, const char *arg2)
{
  ModSnakePyMod *pymod;
  PyThreadState *my_state;
  PyObject *func, *dir_cbdata, *svr_cbdata, *args, *res;
  const char *cres;

  /* Lock */
  cres = mod_snake_setup_internal_directive_call(cmd, per_dir_vdata, &my_state,
						 &dir_cbdata, &svr_cbdata,
						 &func, &pymod);
  if(cres) return cres;

  args = Py_BuildValue("(OOss)", dir_cbdata, svr_cbdata, arg1, arg2); /* INC */
  res = mod_snake_pyeval_callobject(pymod, func, args, SNAKE_HOOK_DIRECTIVE);
  Py_DECREF(args);
  if(res == NULL){
    if(PyErr_Occurred()) PyErr_Print();
    mod_snake_destroy_thread_and_unlock(my_state);  /* Unlock */
    return ap_psprintf(cmd->pool, "mod_snake: Error calling command '%s'",
                      cmd->cmd->name);
  }

  cres = PyString_Check(res) ? ap_pstrdup(cmd->pool, PyString_AsString(res))
                             : NULL;
  Py_DECREF(res);
  mod_snake_destroy_thread_and_unlock(my_state);  /* Unlock */

  return cres;
}


static const 
char *mod_snake_internal_directive_take3(cmd_parms *cmd, void *per_dir_vdata,
                                         const char *arg1, const char *arg2,
                                         const char *arg3)
{
  ModSnakePyMod *pymod;
  PyThreadState *my_state;
  PyObject *func, *dir_cbdata, *svr_cbdata, *args, *res;
  const char *cres;

  /* Lock */
  cres = mod_snake_setup_internal_directive_call(cmd, per_dir_vdata, &my_state,
						 &dir_cbdata, &svr_cbdata,
						 &func, &pymod);
  if(cres) return cres;

  args = Py_BuildValue("(OOsss)", dir_cbdata, svr_cbdata, arg1, arg2, 
		       arg3);  /* Incref */
  res = mod_snake_pyeval_callobject(pymod, func, args, SNAKE_HOOK_DIRECTIVE);
  Py_DECREF(args);
  if(res == NULL){
    if(PyErr_Occurred()) PyErr_Print();
    mod_snake_destroy_thread_and_unlock(my_state);  /* Unlock */
    return ap_psprintf(cmd->pool, "mod_snake: Error calling command '%s'",
                      cmd->cmd->name);
  }

  cres = PyString_Check(res) ? ap_pstrdup(cmd->pool, PyString_AsString(res))
                             : NULL;
  Py_DECREF(res);
  mod_snake_destroy_thread_and_unlock(my_state);  /* Unlock */
  return cres;
}

static AP_CLEANUP_RESTYPE internal_directive_cleanup(void *vdata){
  command_rec *cmd;
  int off = (int)vdata;

  MS_TRACE;

  cmd = (command_rec *)snake_module.cmds + off;
  cmd->cmd_data = NULL;
  return AP_CLEANUP_RESULT(0);
}

/*
 * mod_snake_add_directive:  This routine is a haaaaaaaack.  It allows Python
 *                           modules to create their own directives that
 *                           can be used in Apache configuration.  Each 
 *                           module created adds its own callback information
 *                           to the directory config.  
 *
 * Arguments:                pymod  = Module associated with this directive.
 *                           directive_name = Directive string
 *                           location  = overrides information
 *                           functype  = TAKE1, etc.
 *                           funcptr   = Python func to call on directive
 *                           func_desc = Description of the function
 *
 * Return values:            Returns -1 on error, 0 on success.
 *
 * Better implementation:    It would be much handier if Apache simply allowed
 *                           modules to do dynamic creation of directives.
 */

int mod_snake_add_directive(ModSnakePyMod *pymod,
                            const char *directive_name, int location, 
                            int functype, PyObject *funcptr, 
                            const char *func_desc)
{
  char *loc_directive_name = NULL, *loc_func_desc = NULL;
  ModSnakeIntDirective *intdir;
  command_rec *recs, *find_rec;
  int howmany, presetup = 0;

  /* It may be ok that a directive already exists in the command table.  When
     modules become unloaded, their callback data in the command rec 
     disappears -- they can be replaced by calling back into here */
     
  recs = (command_rec *)snake_module.cmds;
  if((find_rec = (command_rec *)ap_find_command(directive_name, 
						snake_module.cmds)) != NULL)
  {
    presetup = 1;
    if((intdir = find_rec->cmd_data) != NULL)
      return -1;
  } else {
    /* Else, setup a new location for the internal directive to occupy */
    /* First, count how many existing commands we have */
    while(recs->name != NULL)
      recs++;

    howmany = recs - snake_module.cmds + 1; /*  include trailing NULL rec */

    /* Use regular malloc here, as we don't want anything to go away when
       a server config is destroyed. */
    recs = NULL;
    
    loc_directive_name = NULL;
    loc_func_desc      = NULL;
    if((recs = calloc(sizeof(mod_snake_cmds[0]), (howmany + 1))) == NULL ||
       (loc_directive_name = strdup(directive_name)) == NULL ||
       (loc_func_desc = strdup(func_desc)) == NULL)
      {
	/* free(NULL) ok, by ANSI */
	free(recs);
	free(loc_directive_name);
	free(loc_func_desc);
	return -1;
      }
  
    memcpy(recs, snake_module.cmds, sizeof(mod_snake_cmds[0]) * (howmany));
    if(snake_module.cmds != mod_snake_cmds)
      free((command_rec *)snake_module.cmds);  /* Free up old allocs */

    find_rec = recs + howmany - 1;
  }

  Py_INCREF(funcptr);
  intdir          = ap_pcalloc(pymod->ap_p, (sizeof(*intdir)));
  intdir->funcptr = funcptr;
  intdir->pymod   = pymod;

  /* Set the cleanup to NULL out the internal directive client data within
     the command_rec.  We can't store an actual pointer to the data within
     the command rec, because this can change -- instead give an offset
     from the start of the mod_snake command_recs                         */
  ap_register_cleanup(pymod->ap_p, (void *)(find_rec - recs), 
		      internal_directive_cleanup, ap_null_cleanup);
  
  find_rec->cmd_data     = intdir;
  if(!presetup){
    find_rec->name         = loc_directive_name;
    find_rec->req_override = location;
    find_rec->args_how     = functype;
    find_rec->errmsg       = loc_func_desc;
    switch(functype){
    case NO_ARGS:
      find_rec->func = mod_snake_internal_directive_no_args; break;
    case TAKE2:
    case ITERATE2:
    case TAKE12:
      find_rec->func = mod_snake_internal_directive_take2; break;
    case TAKE3:
    case TAKE23:
    case TAKE123:
    case TAKE13:
      find_rec->func = mod_snake_internal_directive_take3; break;
    case TAKE1:
    case ITERATE:
    case FLAG:
    case RAW_ARGS:
    default:
      find_rec->func = mod_snake_internal_directive_take1;
    }
  }

  /* NULL out the last element */
  (find_rec + 1)->name             = NULL;

  snake_module.cmds = recs;
  return 0;
}

/* mod_snake_log_error_unformat:  Log an unformatted error -- this is very
 *                                important for tracebacks and such coming
 *                                from Python code, as meta-characters may
 *                                be in the error string, and throw off our
 *                                varargs.
 *
 * Arguments:                     Similar to ap_log_error, though no format
 */

void 
mod_snake_log_error_unformat(const char *file, int line, int level,
			     ap_status_t status, const server_rec *s,
			     const char *logstr)
{
#ifdef APACHE13
  ap_log_error(file, line, level, s, logstr);
#else
  ap_log_error(file, line, level, status, s, logstr);
#endif
}

 
/* mod_snake_log_error: routine that we can call that will call the
 *                      appropriate ap_log_error function.
 *
 * Arguments:           Same as ap_log_error            
 *
 * Returns:             nothing
 */

void 
mod_snake_log_error(const char *file, int line, int level, ap_status_t status,
                    const server_rec *s, const char *fmt, ...)
{
  va_list args;
  char big_ass_buff[2048];
  
  va_start(args, fmt);
  ap_vsnprintf(big_ass_buff, sizeof(big_ass_buff)-1, fmt, args);
  mod_snake_log_error_unformat(file, line, level, status, s, big_ass_buff);
  va_end(args);
  return;
}


PyObject *
mod_snake_find_cback_data(ModSnakePyMod *pymod, ap_array_header_t *cfgs){
  int i;

  for(i=0; i < cfgs->nelts; i++){
    ModSnakeCallbackData *cbdata = (ModSnakeCallbackData *)cfgs->elts + i;

    if(cbdata->module == pymod)
      return cbdata->data;
  }
  return NULL;
}


PyObject *
mod_snake_find_svr_cback_data(ModSnakePyMod *pymod, ModSnakeSCfg *scfg){
  return mod_snake_find_cback_data(pymod, scfg->svr_config_data);
}

PyObject *
mod_snake_find_dir_cback_data(ModSnakePyMod *pymod, ModSnakeDCfg *dcfg){
  return mod_snake_find_cback_data(pymod, dcfg->dir_config_data);
}
