/* ====================================================================
 * Copyright (c) 2001 Jon Travis.  All rights reserved.
 * See the file COPYING for information on usage and redistribution.
 * ====================================================================
 */
#ifndef APACHE13

#include "mod_snake.h"
#include "mod_snake_buf.h"

struct mod_snake_buf_st {
  char *data;     /* Data contained in the buffer                   */
  int   alloced;  /* Allocated                                      */
  char *endp;     /* Points to the next byte where data can be put  */
};

/*
 * mod_snake_buf_create:  Create a buffer using malloc() off the heap.
 *
 * Arguments:             bufsize = Size of buffer to create
 *
 * Return Values:         Returns NULL on failure, a valid pointer on success.
 */

ModSnakeBuf *mod_snake_buf_create(int bufsize){
  ModSnakeBuf *res;

  if((res = malloc(sizeof(*res))) == NULL)
    return NULL;

  if((res->data = malloc(bufsize)) == NULL){
    free(res);
    return NULL;
  }

  res->alloced = bufsize;
  res->endp    = res->data;
  return res;
}

void mod_snake_buf_destroy(ModSnakeBuf *buf){
  free(buf->data);
  free(buf);
}


/*
 * mod_snake_buf_append:  Append data to the end of a mod_snake buffer.
 *                        Overwriting the end of the buffer is not a 
 *                        suggested practice.. ;-)
 * 
 * Arguments:             buf  = Buffer to append data to
 *                        data = Data to append.  This data is copied.
 *                        len  = Length of data in `data` 
 */

void mod_snake_buf_append(ModSnakeBuf *buf, const char *data, int len){
  memcpy(buf->endp, data, len);
  buf->endp += len;
}

/*
 * mod_snake_buf_size_left:  Get the number of bytes left that can be written
 *                           into the buffer.
 *
 * Arguments:                buf = Buffer to get size left of
 *
 * Return values:            Returns the number of bytes left in the buffer.
 */

int mod_snake_buf_size_left(ModSnakeBuf *buf){
  return buf->alloced;
}

/*
 * mod_snake_buf_bucketize:  Transform a mod_snake buf into a bucket.  The
 *                           input bucket is then destroyed.
 *
 * Return values:            Returns NULL on failure, a new bucket on success.
 *                           On failure, the buffer will not be touched.
 */

apr_bucket *mod_snake_buf_bucketize(ModSnakeBuf *buf){
  apr_bucket *res;

  res = apr_bucket_create_heap(buf->data, buf->endp - buf->data, 0, NULL);
  if(!res)  return NULL;
  free(buf);
  return res;
}


#endif
