/* ====================================================================
 * Copyright (c) 2001 Jon Travis.  All rights reserved.
 * See the file COPYING for information on usage and redistribution.
 * ====================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <unistd.h>
#include <Python.h>
#include <marshal.h>

static
int process_file(const char *infile, const char *outfile, const char *cname){
  FILE *infp = NULL, *outfp = NULL;
  char py_cname[1024];
  char *inbuf, *cp;
  struct stat instat;
  PyObject *code, *codestr = NULL;
  int i;

  if(stat(infile, &instat) == -1){
    fprintf(stderr, "Error stat()ing '%s': %s\n",
	    infile, strerror(errno));
    return -1;
  }

  if((inbuf = malloc(instat.st_size + 1)) == NULL){
    fprintf(stderr, "Error allocating: %ld bytes\n", (long)instat.st_size);
    return -1;
  }

  if((infp = fopen(infile, "r")) == NULL ||
     (outfp = fopen(outfile, "wb")) == NULL)
  {
    fprintf(stderr, "Error opening '%s' for %s: %s\n", 
	    (infp == NULL) ? infile : outfile,
	    (infp == NULL) ? "read" : "write",
	    strerror(errno));
    return -1;
  }

  if(fread(inbuf, 1, instat.st_size, infp) != instat.st_size){
    fprintf(stderr, "Error reading from '%s': %s\n",
	    infile, strerror(errno));
    fclose(infp);
    fclose(outfp);
    free(inbuf);
    return -1;
  }
  fclose(infp);

  inbuf[instat.st_size] = '\0';
  sprintf(py_cname, "<internal %s>", cname);
  if((code = Py_CompileString(inbuf, py_cname, Py_file_input)) == NULL ||
     (codestr = PyMarshal_WriteObjectToString(code)) == NULL)
  {
    fprintf(stderr, "Error compiling python code\n");
    PyErr_Print();
    Py_XDECREF(code);
    Py_XDECREF(codestr);
    fclose(outfp);
    free(inbuf);
    return -1;
  }
  free(inbuf);
  Py_DECREF(code);

  fprintf(outfp, "#include \"snake_precompile.h\"\n\n");
  fprintf(outfp, "static char local%s[] = {\n", cname);
  cp = PyString_AsString(codestr);
  for(i = 0; i < PyString_Size(codestr); i++, cp++){
    fprintf(outfp, "%4d,", (int)*cp);
    if(!((i+1) % 15)) fprintf(outfp, "\n");
  }

  fprintf(outfp, "};\n");
  fprintf(outfp, "const SnakePrecompiledPy %s = {\n", cname);
  fprintf(outfp, "local%s, sizeof(local%s)\n };\n", cname, cname);
  Py_DECREF(codestr);
  fclose(outfp);
  return 0;
}

int main(int argc, char *argv[]){
  Py_Initialize();

  if(argc != 4){
    fprintf(stderr, "Syntax: %s <infile> <outfile> <out_cname>\n", argv[0]);
    return -1;
  }

  return process_file(argv[1], argv[2], argv[3]);
}
