/* ====================================================================
 * Copyright (c) 2001 Jon Travis.  All rights reserved.
 * See the file COPYING for information on usage and redistribution.
 * ====================================================================
 */

#include "mod_snake.h"
#include "mod_snake_buf.h"
#include "mod_snake_bucket.h"

/* This could be const, but Python doesn't dig that.  Python
   may modify this array of data in-place, similar to how Tcl_Eval does,
   so we may want to copy it before evaling it. -- XXX JMT 11-15-00 */
/* ^--- This comment makes no sense .. it isn't even 11-15-00 yet... I must
   have come back in time after writing mod_snake.. ;-)  */

extern SnakePrecompiledPy Mod_Snake_Py;
static int Mod_Snake_Locked = 0;

/*
 * mod_snake_create_thread_and_lock:  Create a new thread state within an
 *                                    interpreter, set that thread current, and
 *                                    acquire the global interpreter lock.
 *
 * Arguments:                         interp = Interpreter within which to 
 *                                             create a new thread.
 *
 * Return values:                     Returns NULL on failure, else a valid
 *                                    threadstate on success.
 */

PyThreadState *mod_snake_create_thread_and_lock(PyInterpreterState *interp){
  PyThreadState *res;
  
  if((res = PyThreadState_New(interp)) == NULL)
    return NULL;

  PyEval_RestoreThread(res);
  Mod_Snake_Locked = 1;
  return res;
}

int mod_snake_is_locked(){
  return Mod_Snake_Locked;
}


/*
 * mod_snake_destroy_thread_and_unlock:  Destroy a thread state, and release
 *                                       the interpreter lock
 *
 * Arguments:                            tstate = Threadstate to destroy
 */

void mod_snake_destroy_thread_and_unlock(PyThreadState *tstate){
  PyThreadState_Clear(tstate);
  Mod_Snake_Locked = 0;
  PyEval_SaveThread();
  PyThreadState_Delete(tstate);
}

/*
 * initmod_internal:  Load a statically compiled in module into the main
 *                    interp so that other modules have access to it.  
 *                    This sets up a dummy module and uses its dict to 
 *                    load all the proper info.  The Python global lock
 *                    MUST be acquired before calling this function.
 *
 * Arguments:         modname = Name to assign the module for 'import blah'
 *                    modstr  = Precompiled module information.
 *
 * Return values:  Returns -1 on failure, 0 on success.
 */

int mod_snake_initmod_internal(char *modname, SnakePrecompiledPy *modstr){
  PyObject *pmod, *pdict;

  if((pmod = mod_snake_exec_precompile(modname, modstr)) == NULL)
    return -1;
  
  pdict = PyModule_GetDict(pmod);   /* No Incref */
  PyDict_SetItemString(pdict, "__dummy__", Py_None);  /* Alert Python to dummy
                                                         module status */
  return 0;
}


/*
 * mod_snake_setup_syspath:  Setup the search path in the sys module.  If
 *                           any of the additions to the list fails, the
 *                           state of the list is undefined.  
 *
 * Arguments:                moddirs = array of module dirs to add to sys.path
 * 
 * Return values:            Returns -1 on failure, 0 on success.
 */

int mod_snake_setup_syspath(ap_array_header_t *moddirs){
  int i, err;
  PyObject *sysmod, *sysdict, *syspath, *str;

  /* Attempt to add search directories to sys.path.  The sys.path will
     be shared upon deletions and re-imports from other modules.  */
  
  if((sysmod = PyImport_ImportModule("sys")) == NULL)  /* Incref'd */
    return -1;

  if((sysdict = PyModule_GetDict(sysmod)) == NULL ||
     (syspath = PyDict_GetItemString(sysdict, "path")) == NULL){
    Py_DECREF(sysmod);
    return -1;
  }

  for(i=0; i< moddirs->nelts; i++){
    str = PyString_FromString(*((char **)moddirs->elts + i));
    err = PyList_Append(syspath, str);
    Py_DECREF(str);
    if(err){
      Py_DECREF(sysmod);
      return -1;
    }
  }

  Py_DECREF(sysmod);
  return 0;
}

/*
 * mod_snake_import_module:  Import a module into the server.  The module is 
 *                           imported into the current thread/interp.  On 
 *                           failure, pymod will be:
 *                           'the same as it ever was' (TM)
 *
 * Arguments:                 module = Module to load
 *                            s      = Server record to use for logging
 * 
 * Side Notes:                The Python interpreter lock MUST be held before
 *                            calling this function.
 *
 * Return values:             Returns -1 on failure, 0 on success.  
 */

int mod_snake_import_module(ModSnakePyMod *pymod, server_rec *s)
{
  PyObject *module, *reg_class, *py_modpy, *pargs, *mod_obj;
  char buf[1024], *cp;

  strncpy(buf, pymod->name, sizeof(buf) - 1);
  buf[sizeof(buf)-1] = '\0';
  if((cp = strchr(buf, '.')) == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, s,
                 "mod_snake: module must be in the form of module.classname");
    return -1;
  }

  *cp = '\0';
  if((module = PyImport_ImportModule(buf)) == NULL){ /* Incref*/
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, s,
                 "mod_snake: error loading module '%s'", pymod->name);
    if(PyErr_Occurred()) PyErr_Print();
    return -1;      
  }

  /* Load the registration class */
  if((reg_class = PyObject_GetAttrString(module, cp + 1)) == NULL){ /* Incref*/
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, s,
                 "mod_snake: registration class '%s' not found in %s",
                 cp + 1, buf);
    Py_DECREF(module);
    if(PyErr_Occurred()) PyErr_Print();
    return -1;
  }

  /* Instantiate the class */
  py_modpy = mod_snake_get_modsnakepymod_shadowclass(pymod);
  pargs = Py_BuildValue("(O)", py_modpy); /* Incref */
  Py_DECREF(py_modpy);

  pymod->current_hook = SNAKE_HOOK_INIT;
  if((mod_obj = PyEval_CallObject(reg_class, pargs)) == NULL){ /* Incref */
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, s,
                 "mod_snake: Error instantiating class '%s'", cp + 1);
    PyErr_Print();
    Py_DECREF(pargs);
    Py_DECREF(reg_class);
    Py_DECREF(module);
    pymod->current_hook = SNAKE_HOOK_NO_HOOK;
    return -1;
  }
  pymod->current_hook = SNAKE_HOOK_NO_HOOK;

  Py_DECREF(pargs);
  Py_DECREF(reg_class);
  
  pymod->mod_obj = mod_obj;
  pymod->module  = module;
  return 0;
}


/*
 * mod_snake_load_profiler: Load the Python profiler module into an interpreter
 *                          and instantiate a Profile object for profiling
 *                          within the PyMod.  Before calling this function,
 *                          the interpreter must be set to that of the
 *                          module, and the global interpreter lock must be 
 *                          held.
 *
 * Return values:           Returns NULL on failure, else a PyObject 
 *                          instantiation of the profile.Profile class with
 *                          a new INCREF
 */

static PyObject *mod_snake_load_profiler(void){
  PyObject *profile_module, *profile_class, *profile_obj;

  if((profile_module = PyImport_ImportModule("profile2")) == NULL) /* Incref */
    return NULL;

  if((profile_class = PyObject_GetAttrString(profile_module, 
                                             "Profile")) == NULL) /* Incref */
  {
    Py_DECREF(profile_module);
    return NULL;
  }
  
  profile_obj = PyEval_CallObject(profile_class, NULL);  /* Incref */
  Py_DECREF(profile_class);
  Py_DECREF(profile_module);
  return profile_obj;
}

static 
PyObject *mod_snake_load_profile_func(PyObject *profile_obj){
  return PyObject_GetAttrString(profile_obj, "runcall");  /* INCREF */
}



/*
 * mod_snake_load_module:   Setup an interpreter, system path and
 *                          import internal modules into a server mod
 *                          that we're importing
 *
 * Arguments:               pymod   = Module to setup
 *                          syspath = System path to add
 *                          s       = Server for error reporting
 *                      
 *
 * Return values:           returns -1 on failure, 0 on success.
 */

int mod_snake_load_module(ModSnakePyMod *pymod, ap_array_header_t *syspath,
                          server_rec *s)
{
  PyThreadState *thr, *old_state;
  PyObject *profile_obj = NULL, *profile_func = NULL;

  /* Setup a new interpreter for this module */
  if((thr = Py_NewInterpreter()) == NULL)
    return -1;
  
  /* Swap into the new interp so that all subsequent operations
     are performed within it */
  old_state = PyThreadState_Swap(thr);
  if(mod_snake_setup_syspath(syspath) == -1){
    Py_EndInterpreter(thr);
    PyThreadState_Swap(old_state);
    return -1;
  }
  
  /* Call our SWIG loader for wrapped mod_snake module */
  initmod_snakec();

  if(mod_snake_initmod_internal("mod_snake", &Mod_Snake_Py) == -1){
    Py_EndInterpreter(thr);
    PyThreadState_Swap(old_state);
    return -1;
  }

  if(pymod->profile_info.profile && 
     ((profile_obj = mod_snake_load_profiler()) == NULL /* INCREF both */
      || (profile_func = mod_snake_load_profile_func(profile_obj)) == NULL))
  {
    Py_XDECREF(profile_obj);
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, s,
                        "mod_snake: Error loading profiler!");
    if(PyErr_Occurred())
      PyErr_Print();
    Py_EndInterpreter(thr);
    PyThreadState_Swap(old_state);
    return -1;
  }

  if(mod_snake_import_module(pymod, s) == -1){
    Py_XDECREF(profile_obj);  /* pymod->profile_info.profile may == 0 */
    Py_XDECREF(profile_func);
    Py_EndInterpreter(thr);
    PyThreadState_Swap(old_state);
    return -1;
  }

  pymod->interp                    = thr->interp;
  pymod->main_state                = thr;
  pymod->valid                     = 1;
  pymod->profile_info.profile_obj  = profile_obj;
  pymod->profile_info.profile_func = profile_func;
  PyThreadState_Swap(old_state);
  return 0;
}

static struct hook_mapping_st {
  const char *name;
  int hook_type;
} HookNameMappings[SNAKE_HOOK_NUM_HOOKS] = {
  { "create_dir_config", SNAKE_HOOK_BOFFO },/* SNAKE_HOOK_CREATE_DIR_CONFIG  */
  { "merge_dir_config",  SNAKE_HOOK_BOFFO },/* SNAKE_HOOK_MERGE_DIR_CONFIG   */
  { "create_svr_config", SNAKE_HOOK_BOFFO },/* SNAKE_HOOK_CREATE_SVR_CONFIG  */
  { "merge_svr_config",  SNAKE_HOOK_BOFFO },/* SNAKE_HOOK_MERGE_SVR_CONFIG   */
  { "post_config",       SNAKE_HOOK_VOID  },/* SNAKE_HOOK_POST_CONFIG        */
  { "open_logs",         SNAKE_HOOK_VOID  },/* SNAKE_HOOK_OPEN_LOGS          */
  { "header_parser",     SNAKE_HOOK_ALL   },/* SNAKE_HOOK_HEADER_PARSER      */
  { "post_read_request", SNAKE_HOOK_ALL   },/* SNAKE_HOOK_POST_READ_REQUEST  */
  { "log_transaction",   SNAKE_HOOK_ALL   },/* SNAKE_HOOK_LOG_TRANSACTION    */
  { "translate_name",    SNAKE_HOOK_FIRST },/* SNAKE_HOOK_TRANSLATE_NAME     */
  { "check_user_id",     SNAKE_HOOK_FIRST },/* SNAKE_HOOK_CHECK_USER_ID      */
  { "fixups",            SNAKE_HOOK_ALL   },/* SNAKE_HOOK_FIXUPS             */
  { "type_checker",      SNAKE_HOOK_FIRST },/* SNAKE_HOOK_TYPE_CHECKER       */
  { "access_checker",    SNAKE_HOOK_ALL   },/* SNAKE_HOOK_ACCESS_CHECKER     */
  { "auth_checker",      SNAKE_HOOK_FIRST },/* SNAKE_HOOK_AUTH_CHECKER       */
  { "content_handler",   SNAKE_HOOK_FIRST },/* SNAKE_HOOK_CONTENT_HANDLER    */
#ifndef APACHE13
  { "pre_connection",    SNAKE_HOOK_ALL   },/* SNAKE_HOOK_PRE_CONNECTION     */
  { "process_connection",SNAKE_HOOK_FIRST },/* SNAKE_HOOK_PROCESS_CONNECTION */
  { "insert_filter",     SNAKE_HOOK_VOID  },/* SNAKE_HOOK_INSERT_FILTER      */
  { "run_filter",        SNAKE_HOOK_BOFFO },/* SNAKE_HOOK_RUN_FILTER         */
#endif
};


int mod_snake_hooknum_to_type(int hookno){
  if(hookno < 0 || hookno >= SNAKE_HOOK_NUM_HOOKS)
    return SNAKE_HOOK_BOFFO;
  
  return HookNameMappings[hookno].hook_type;
}
   
int mod_snake_hookname_to_num(const char *hookname){
  int i;

  for(i=0; i< SNAKE_HOOK_NUM_HOOKS; i++)
    if(!strcmp(hookname, HookNameMappings[i].name)) return i;

  return -1;
}

const char *mod_snake_hooknum_to_name(int hookno){
  if(hookno < 0 || hookno >= SNAKE_HOOK_NUM_HOOKS)
    return "Unknown hook type";

  return HookNameMappings[hookno].name;
}

#ifndef APACHE13
int mod_snake_run_filter_hook(ModSnakePyMod *pymod, PyObject *func,
			      PyObject *per_dir_data,
			      PyObject *per_svr_data,
			      ap_filter_t *f, apr_bucket_brigade *brigade,
			      int do_lock)
{
  server_rec *s = mod_snake_get_main_server();
  PyObject *pargs, *pres, *pfilter, *pbrigade;
  PyThreadState *tstate = NULL;
  int res;

  /* Lock */
  if(do_lock)
    if((tstate = mod_snake_create_thread_and_lock(pymod->interp)) == NULL){
      mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, s,
			  "mod_snake: Failed to create a new threadstate");
      return HTTP_INTERNAL_SERVER_ERROR;
    }
  
  if((pfilter = mod_snake_get_ap_filter_t_shadowclass(f)) == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, s,
                 "mod_snake: (%s) Error converting internal ap_filter_t rep",
		 pymod->name);
    if(PyErr_Occurred()) PyErr_Print();
    if(do_lock) mod_snake_destroy_thread_and_unlock(tstate);  /* Unlock */
    return HTTP_INTERNAL_SERVER_ERROR;
  }
  
  if((pbrigade = mod_snake_get_apr_bucket_brigade_shadowclass(brigade))== NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, s,
            "mod_snake: (%s) Error converting internal apr_bucket_brigade rep",
	    pymod->name);
    if(PyErr_Occurred()) PyErr_Print();
    if(do_lock) mod_snake_destroy_thread_and_unlock(tstate);  /* Unlock */
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  mod_snake_set_curpymod(f->r->connection, f->r, pymod);
  /* pargs, pfilter, pbrigade incref */
  pargs = Py_BuildValue("(OOOO)", per_dir_data, per_svr_data, pfilter, 
			pbrigade);
  Py_DECREF(pfilter);                   
  Py_DECREF(pbrigade);

  pres = mod_snake_pyeval_callobject(pymod, func, pargs, 
				     SNAKE_HOOK_RUN_FILTER); /*Incref*/
  Py_DECREF(pargs);
  if(pres == NULL || !PyInt_Check(pres)){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, s,
                "mod_snake: (%s) Failed to call %s hook",
                pymod->name, mod_snake_hooknum_to_name(SNAKE_HOOK_RUN_FILTER));
    if(!PyErr_Occurred() && !PyInt_Check(pres)){
      PyErr_SetString(PyExc_TypeError, "Return value not an int");
      PyErr_Print();
    } else if(PyErr_Occurred()) PyErr_Print();
    if(do_lock) mod_snake_destroy_thread_and_unlock(tstate);   /* Unlock */
    return HTTP_INTERNAL_SERVER_ERROR;
  }
  res = PyInt_AsLong(pres);
  if(do_lock) mod_snake_destroy_thread_and_unlock(tstate);     /* Unlock */
  return res;
}
#endif

PyObject * 
mod_snake_run_hook_create_dir_config(ModSnakePyMod *pymod, 
				     ModSnakeDCfg *dir_cfg,
				     PyObject *callback)
{
  PyObject *pargs, *pres;
  PyThreadState *tstate;
  server_rec *s = mod_snake_get_main_server();

  /* Lock */
  if((tstate = mod_snake_create_thread_and_lock(pymod->interp)) == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, s,
                 "mod_snake: Failed to create a new threadstate");
    return NULL;
  }
  
  pargs = Py_BuildValue("(s)", dir_cfg->path);             /* pargs incref */
  pres = mod_snake_pyeval_callobject(pymod, callback, pargs, 
                                     SNAKE_HOOK_CREATE_DIR_CONFIG); /* Incref*/
  Py_DECREF(pargs);
  if(pres == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, s,
                 "mod_snake: (%s) Failed to call create_dir_config hook",
                 pymod->name);
    if(PyErr_Occurred()) PyErr_Print();
    mod_snake_destroy_thread_and_unlock(tstate);
    return NULL;
  }
  mod_snake_destroy_thread_and_unlock(tstate);
  return pres;
}

PyObject * 
mod_snake_run_hook_create_svr_config(ModSnakePyMod *pymod, server_rec *s,
                                     ModSnakeSCfg *svr_cfg,
                                     PyObject *callback)
{
  PyObject *pargs, *pres, *svr_shadowclass;
  PyThreadState *tstate;

  /* Lock */
  if((tstate = mod_snake_create_thread_and_lock(pymod->interp)) == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, s,
                 "mod_snake: Failed to create a new threadstate");
    return NULL;
  }
  
  if((svr_shadowclass = mod_snake_get_server_rec_shadowclass(s)) == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, s,
                 "mod_snake: Failed to create a server shadowclass object");
    mod_snake_destroy_thread_and_unlock(tstate);
    return NULL;
  }

  pargs = Py_BuildValue("(O)", svr_shadowclass);         /* pargs incref */
  Py_DECREF(svr_shadowclass);
  pres = mod_snake_pyeval_callobject(pymod, callback, pargs, 
                                     SNAKE_HOOK_CREATE_DIR_CONFIG); /* Incref*/
  Py_DECREF(pargs);
  if(pres == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, s,
                 "mod_snake: (%s) Failed to call create_dir_config hook",
                 pymod->name);
    if(PyErr_Occurred()) PyErr_Print();
    mod_snake_destroy_thread_and_unlock(tstate);
    return NULL;
  }
  mod_snake_destroy_thread_and_unlock(tstate);
  return pres;
}

static PyObject * 
mod_snake_run_hook_merge_config(ModSnakePyMod *pymod, 
				PyObject *base_cdata, PyObject *new_cdata,
				PyObject *callback, int hookno)
{
  PyObject *pargs, *pres;
  PyThreadState *tstate;
  server_rec *s = mod_snake_get_main_server();

  /* Lock */
  if((tstate = mod_snake_create_thread_and_lock(pymod->interp)) == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, s,
                 "mod_snake: Failed to create a new threadstate");
    return NULL;
  }
  
  pargs = Py_BuildValue("(OO)", base_cdata, new_cdata);    /* pargs incref */
  pres = mod_snake_pyeval_callobject(pymod, callback, pargs, hookno); /* INC */
  Py_DECREF(pargs);
  if(pres == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, s,
                 "mod_snake: (%s) Failed to call %s hook",
                 pymod->name, mod_snake_hooknum_to_name(hookno));
    if(PyErr_Occurred()) PyErr_Print();
    mod_snake_destroy_thread_and_unlock(tstate);
    return NULL;
  }
  mod_snake_destroy_thread_and_unlock(tstate);
  return pres;
}

PyObject * 
mod_snake_run_hook_merge_dir_config(ModSnakePyMod *pymod, PyObject *base, 
				    PyObject *newc, PyObject *callback)
{
  return mod_snake_run_hook_merge_config(pymod, base, newc, callback,
					 SNAKE_HOOK_MERGE_DIR_CONFIG);
}

PyObject * 
mod_snake_run_hook_merge_svr_config(ModSnakePyMod *pymod, PyObject *base,
				    PyObject *newc, PyObject *callback)
{
  return mod_snake_run_hook_merge_config(pymod, base, newc, callback,
					 SNAKE_HOOK_MERGE_SVR_CONFIG);
}

int mod_snake_run_generic_conn_rec(ModSnakePyMod *pymod, conn_rec *c,
				   PyObject *per_svr_data,
				   PyObject *callback, int cb_num)
{
  PyObject *pargs, *pres, *pconn_rec;
  PyThreadState *tstate;
  int res;

  /* Lock */
  if((tstate = mod_snake_create_thread_and_lock(pymod->interp)) == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0,c->base_server,
                 "mod_snake: Failed to create a new threadstate");
    return HTTP_INTERNAL_SERVER_ERROR;
  }
  
  if((pconn_rec = mod_snake_get_conn_rec_shadowclass(c)) == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, c->base_server,
                 "mod_snake: (%s) Error converting internal conn_rec rep",
                 pymod->name);
    if(PyErr_Occurred()) PyErr_Print();
    mod_snake_destroy_thread_and_unlock(tstate);  /* Unlock */
    return HTTP_INTERNAL_SERVER_ERROR;
  }
  
  mod_snake_set_curpymod(c, NULL, pymod);
  /* pargs, pconn_rec, pconn_data incref */
  pargs = Py_BuildValue("(OO)", per_svr_data, pconn_rec);
  Py_DECREF(pconn_rec);                   

  pres = mod_snake_pyeval_callobject(pymod, callback,pargs, cb_num); /*Incref*/
  Py_DECREF(pargs);
  if(pres == NULL || !PyInt_Check(pres)){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0,c->base_server,
                 "mod_snake: (%s) Failed to call %s hook",
                 pymod->name, mod_snake_hooknum_to_name(cb_num));
    if(!PyErr_Occurred() && !PyInt_Check(pres)){
      PyErr_SetString(PyExc_TypeError, "Return value not an int");
      PyErr_Print();
    } else if(PyErr_Occurred()) PyErr_Print();
    mod_snake_destroy_thread_and_unlock(tstate);   /* Unlock */
    return HTTP_INTERNAL_SERVER_ERROR;
  }
  res = PyInt_AsLong(pres);
  mod_snake_destroy_thread_and_unlock(tstate);     /* Unlock */
  return res;
}


int mod_snake_run_generic_request_rec(ModSnakePyMod *pymod, request_rec *r,
				      PyObject *per_dir_data,
				      PyObject *per_svr_data,
                                      PyObject *callback, int cb_num)
{
  PyObject *pargs, *pres, *prequest_rec;
  PyThreadState *tstate;
  ModSnakeReqCfg *rcfg = mod_snake_get_req_cfg(r);
  int res;

  /* Lock */
  if((tstate = mod_snake_create_thread_and_lock(pymod->interp)) == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, r->server,
                 "mod_snake: Failed to create a new threadstate");
    return HTTP_INTERNAL_SERVER_ERROR;
  }
  
  if((prequest_rec = mod_snake_get_request_rec_shadowclass(r)) == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, r->server,
                 "mod_snake: (%s) Error converting internal request_rec rep",
                 pymod->name);
    if(PyErr_Occurred()) PyErr_Print();
    mod_snake_destroy_thread_and_unlock(tstate);  /* Unlock */
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  mod_snake_set_curpymod(r->connection, r, pymod);
  pargs = Py_BuildValue("(OOO)", per_dir_data, per_svr_data, prequest_rec); 
  /* pargs + pcb_data + prequest_rec incref */
  Py_DECREF(prequest_rec);                   
  pres = mod_snake_pyeval_callobject(pymod, callback,pargs, cb_num); /*Incref*/
  Py_DECREF(pargs);
  if(pres == NULL || !PyInt_Check(pres)){
    /* XXX -- Prehaps cleanup the rwrite buffer here? */
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, r->server,
                 "mod_snake: (%s) Failed to call %s hook",
                 pymod->name, mod_snake_hooknum_to_name(cb_num));
    if(!PyErr_Occurred() && !PyInt_Check(pres)){
      PyErr_SetString(PyExc_TypeError, "Return value not an int");
      PyErr_Print();
    } else if(PyErr_Occurred()) 
      PyErr_Print();
    mod_snake_destroy_thread_and_unlock(tstate);   /* Unlock */
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  res = PyInt_AsLong(pres);
#ifndef APACHE13
  /* If this was a generate-content callback, then we need to check if the
     module was using the old-school rwrite() methods.  If so, we need to 
     ensure that everything is in the bucket brigade to be passed down, and
     that we fire off an EOS to flushy flushy */

  if(cb_num == SNAKE_HOOK_CONTENT_HANDLER && rcfg->bufused){
    apr_bucket_brigade *brigade;
    apr_bucket *bucket;

    brigade = apr_brigade_create(r->pool);
    if(rcfg->curbuf){
      if(res == HTTP_INTERNAL_SERVER_ERROR){
	mod_snake_buf_destroy(rcfg->curbuf);
	mod_snake_destroy_thread_and_unlock(tstate);
	return res;
      }
      if((bucket = mod_snake_buf_bucketize(rcfg->curbuf)) == NULL){
	mod_snake_buf_destroy(rcfg->curbuf);
	PyErr_SetString(PyExc_MemoryError, "failed to bucketize buffer");
	mod_snake_destroy_thread_and_unlock(tstate);
	return -1;
      }
      APR_BRIGADE_INSERT_TAIL(brigade, bucket);
    }
    bucket = apr_bucket_create_eos();
    APR_BRIGADE_INSERT_TAIL(brigade, bucket);
    ap_pass_brigade(r->output_filters, brigade);
  }

#endif

  mod_snake_destroy_thread_and_unlock(tstate);     /* Unlock */
  return res;
}

void mod_snake_run_hook_post_config(ModSnakePyMod *pymod, ModSnakeSCfg *scfg){
  PyObject *pargs, *pres, *per_svr_data;
  PyThreadState *tstate;

  /* Lock */
  if((tstate = mod_snake_create_thread_and_lock(pymod->interp)) == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 
			0, mod_snake_get_main_server(),
			"mod_snake: Failed to create a new threadstate");
    return;
  }
  
  if(!(per_svr_data = mod_snake_find_svr_cback_data(pymod, scfg)))
    per_svr_data = Py_None;

  pargs = Py_BuildValue("(O)", per_svr_data); /* pargs INC */
  pres =mod_snake_pyeval_callobject(pymod,pymod->hooks[SNAKE_HOOK_POST_CONFIG],
                                     pargs,SNAKE_HOOK_POST_CONFIG);/* Incref */
  Py_DECREF(pargs);
  if(pres == NULL){
    if(PyErr_Occurred())
      PyErr_Print();
  } else
    Py_DECREF(pres);  /* Discard the result */
  mod_snake_destroy_thread_and_unlock(tstate);     /* Unlock */
  return;
}

void mod_snake_run_hook_open_logs(ModSnakePyMod *pymod, ModSnakeSCfg *scfg){
  PyObject *pargs, *pres, *ppymod, *per_svr_data;
  PyThreadState *tstate;

  /* Lock */
  if((tstate = mod_snake_create_thread_and_lock(pymod->interp)) == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 
			0, mod_snake_get_main_server(),
			"mod_snake: Failed to create a new threadstate");
    return;
  }
  
  ppymod = mod_snake_get_modsnakepymod_shadowclass(pymod);
  if(!(per_svr_data = mod_snake_find_svr_cback_data(pymod, scfg)))
    per_svr_data = Py_None;

  pargs = Py_BuildValue("(OO)", per_svr_data, ppymod); /* pargs + ppymod INC */
  Py_DECREF(ppymod);
  pres = mod_snake_pyeval_callobject(pymod, pymod->hooks[SNAKE_HOOK_OPEN_LOGS],
                                     pargs, SNAKE_HOOK_OPEN_LOGS); /* Incref */
  Py_DECREF(pargs);
  if(pres == NULL){
    if(PyErr_Occurred())
      PyErr_Print();
  } else
    Py_DECREF(pres);  /* Discard the result */
  mod_snake_destroy_thread_and_unlock(tstate);     /* Unlock */
  return;
}


PyObject *mod_snake_pyeval_callobject(ModSnakePyMod *pymod, PyObject *func, 
                                      PyObject *args, int hook_num)
{
  PyObject *res;

  pymod->current_hook = hook_num;
  if(pymod->profile_info.profile){
    /* Since the runcall function takes a varags we have to pull the
       tuple layer back one level for the 'args'.  I.e. we have to create
       a new tuple with all the old args, but our new func as the first
       argument.  We could use TupleResize to be much more efficient, but
       we might as well be safe (we can afford to do slow things since
       we are profiling anyway) */
    PyObject *new_pargs;
    int i;

    if((new_pargs = PyTuple_New(PyTuple_Size(args) + 1)) == NULL)
      return NULL;

    Py_INCREF(func);           /* Incref func      */
    PyTuple_SetItem(new_pargs, 0, func);
    for(i=0; i< PyTuple_Size(args); i++){
      /* INCREF each of the passed in args */
      Py_INCREF(PyTuple_GET_ITEM(args, i));
      PyTuple_SET_ITEM(new_pargs, i + 1, PyTuple_GET_ITEM(args, i)); 
    }
    
    res = PyEval_CallObject(pymod->profile_info.profile_func,  
                            new_pargs);   /* Incref res */
    Py_DECREF(new_pargs); 
  } else {
    res = PyEval_CallObject(func, args);  /* Incref res */
  }
  pymod->current_hook = SNAKE_HOOK_NO_HOOK;
  return res;
}

/*
 * mod_snake_tls_get:  Get a value, from thread-local storage, based on
 *                     an key.  
 *
 * Side Notes:         The Python interpreter lock MUST be held before
 *                     calling this function.  The TLS within the
 *                     current thread, will be the one used for storage.
 *
 * Arguments:          key = Key to use to retrieve a value entry.
 *
 * Return Values:      Returns NULL on failure, else the value stored
 *                     in the dict, which matches the key.  The returned
 *                     value will not be INCREF'd
 */

PyObject *mod_snake_tls_get(PyObject *key){
  PyObject *pdict = PyThreadState_GetDict();

  if(!pdict)
    return NULL;

  return PyDict_GetItem(pdict, key);
}


/*
 * mod_snake_tls_set:  Set a key-value pair within thread-local storage.
 *                     This is a somewhat 'safe' area to put global data.
 *
 * Side Notes:         The Python interpreter lock MUST be held before
 *                     calling this function.  The TLS within the
 *                     current thread, will be the one used for storage.
 *
 * Arguments:          key = PyObject to use as the key into the
 *                           dictionary.  This value must be hashable,
 *                           or an error will occur.
 *
 *                     val = PyObject to store as the value in the dict.
 *
 * Return values:      Returns -1 on error, 0 on success.  On error, a
 *                     Python error value will exist.  On success, both
 *                     key and value objects will be INCREF'd
 */

int mod_snake_tls_set(PyObject *key, PyObject *val){
  PyObject *pdict = PyThreadState_GetDict();

  if(!pdict)
    return -1;

  if(PyDict_SetItem(pdict, key, val) == -1)
    return -1;
  return 0;
}

void 
mod_snake_set_curpymod(conn_rec *conn, request_rec *req, ModSnakePyMod *pymod){
  ModSnakeConnCfg *cfg = mod_snake_get_conn_cfg(conn, req);
  cfg->pymod = pymod;
}

ModSnakePyMod *mod_snake_get_curpymod(conn_rec *conn, request_rec *req){
  ModSnakeConnCfg *cfg;

  if((cfg = mod_snake_get_conn_cfg(conn, req)) == NULL)
    return NULL;
  return cfg->pymod;
}
